// ****************************************************************************
//
// avr-sizex
// Copyright 2002 Brian Low
//
// DESCRIPTION
// Calculates the amount of flash, sram and eeprom memory used on an
// AVR device based on an ELF file.
//
//
// LICENSE
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
// 02111-1307  USA
//
//
// HISTORY
//	2002-Nov-08		v0.1	- first test release
//	2002-Nov-09		v0.2	- remove external dependency on cygwin
//							  (now using mingw)
//	2002-Nov-12		v0.3	- fixed random "unable to find .text section" error
//							- fixed --long not recognized
//							- added --top option
//							- display usage if no cmd line args passed
//
//
// BUGS/TODO
//	- can we estimate max stack size?
//	- how do we handle ATtiny devices with no sram?
//
// (tab size for this file: 4)
// ****************************************************************************

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include "getopt.h"


// ------------------------------------------------------------------
// Version of this utility, increase after every release
// ------------------------------------------------------------------
#define VERSION		"0.3"


// ------------------------------------------------------------------
// Magic number identifying a file as an ELF file
// ------------------------------------------------------------------
const char ELF_IDENT[16] = {
	0x7f, 0x45, 0x4c, 0x46,
	0x01, 0x01, 0x01, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00 };

// ------------------------------------------------------------------
// Definition for ELF file header
// ------------------------------------------------------------------
typedef struct
{
  char	e_ident[16];		// magic number identifying file as ELF
  short int e_type;			// object file typ
  short int e_machine;		// target architecture
  int   e_version;			// ELF version
  char *e_entry;  			// starting virtual address
  int   e_phoff;			// file offset to program headers
  int   e_shoff;			// file offset to section headers
  int   e_flags;			// processor-specific flags
  short int e_ehsize;		// ELF header size
  short int e_phentsize;	// size of each program header entry
  short int e_phnum;		// number of program headers
  short int e_shentsize;	// size of each section header entry
  short int e_shnum;		// number of section headers
  short int e_shstrndx;		// index to the "section header string" section
} Elf32_Header;


// ------------------------------------------------------------------
// Definition for ELF section headers
// ------------------------------------------------------------------
typedef struct
{
  int sh_name;				// offset into section header strings table
  int sh_type;				// type of the section
  int sh_flags;				// section-specific flags
  int sh_addr;				// memory location of section
  int sh_offset;			// file offset to section
  int sh_size;				// size of section
  int sh_link;				// section type dependent
  int sh_info;				// extra information
  int sh_addralign;			// address alignment constraint
  int sh_entsize;			// ?
} Elf32_Section_Header;


// ------------------------------------------------------------------
// Definition for ELF symbol
// ------------------------------------------------------------------
typedef struct {
  int			st_name;	// name of symbol
  unsigned int	st_value;
  int			st_size;	// size of symbol
  unsigned char	st_info;
  unsigned char	st_other;
  short int		st_shndx;
} Elf32_Symbol;


// ------------------------------------------------------------------
// This info is needed when parsing the symbol table */
// ------------------------------------------------------------------
#define STB_LOCAL  0
#define STB_GLOBAL 1
#define STB_WEAK   2

#define STT_NOTYPE  0
#define STT_OBJECT  1
#define STT_FUNC    2
#define STT_SECTION 3
#define STT_FILE    4

#define ELF32_ST_BIND(x) ((x) >> 4)
#define ELF32_ST_TYPE(x) (((unsigned int) x) & 0x0f)
#define ELF32_ST_INFO(bind,type) (((bind) << 4) | ((type) & 0x0f))


// ------------------------------------------------------------------
// Definition of command line options this utlity accepts
// Used by getopt_long() in getopt.h
// ------------------------------------------------------------------
static struct option longopts[] = {
	{ "mcu",		required_argument,	0, 		'm' },
	{ "flash",		required_argument,  0,		'f' },
	{ "sram", 		required_argument, 	0, 		's' },
	{ "eeprom",		required_argument,	0,		'e' },
	{ "long", 		no_argument,		0,		'l' },
	{ "top", 		required_argument,	0,		't' },
	{ "nowarn",		no_argument,		0, 		'w'	},
	{ "noerror", 	no_argument,		0,		'o' },
	{ "version", 	no_argument,		0,		'v' },
	{ "help",		no_argument,		0,		'h' },
	{ "devices",	no_argument, 		0,		'd' }
};


// ------------------------------------------------------------------
// Definition for describing Atmel AVR devices
// ------------------------------------------------------------------
typedef struct
{
	char	*name;
	int		flash;
	int		sram;
	int		eeprom;
} device;

// ------------------------------------------------------------------
// Device description list
// Note: the - (minus) in the 2nd column is subtracting the
// space reserved for the boot loader. The size of the
// boot area is configurable so the default was used.
// ------------------------------------------------------------------
static device devices[] = {
	{ "ATtiny11",		1024 	-    0,		0,		0		},
	{ "ATtiny11L",		1024 	-    0,		0,		0		},
	{ "ATtiny12",		1024	-    0,		0,		64		},
	{ "ATtiny12L",		1024	-    0,		0,		64		},
	{ "ATtiny12V",		1024	-    0,		0,		64		},
	{ "ATtiny15",		1024	-    0,		0,		64		},
	{ "ATtiny15L",		1024	-    0,		0,		64		},
	{ "ATtiny22",		2048	-    0,		128,	128		},
	{ "ATtiny22L",		2048	-    0,		128,	128		},
	{ "ATtiny26",		2048	-    0,		128,	128		},
	{ "ATtiny26L",		2048	-    0,		128,	128		},
	{ "ATtiny28",		2048	-    0,		0,		0		},
	{ "ATtiny28L",		2048	-    0,		0,		0		},
	{ "ATtiny28V",		2048	-    0,		0,		0		},
	{ "AT90S1200",		1024	-    0,		0,		64		},
	{ "AT90S2313",		2048	-    0,		128,	128		},
	{ "AT90S2343",		2048	-    0,		128,	128		},
	{ "AT90S2333",		2048	-    0,		128,	128		},
	{ "AT90S2323",		2048	-    0,		128,	128		},
	{ "AT90S4433",		4096	-    0,		128,	256		},
	{ "AT90S4414",		4096	-    0,		256,	256		},
	{ "AT90S4434",		4096	-    0,		256,	256		},
	{ "AT90LS2333",		2048	-    0,		128,	128		},
	{ "AT90LS2343",		2048	-    0,		128,	128		},
	{ "AT90LS2323",		2048	-    0,		128,	128		},
	{ "AT90LS4433",		4096	-    0,		128,	256		},
	{ "AT90LS4434",		4096	-    0,		256,	256		},
	{ "AT90LS8535",		8192	-    0,		512,	512		},
	{ "AT90S8515",		8192	-    0,		512,	512		},
	{ "AT90S8535",		8192	-    0,		512,	512		},
	{ "AT90C8534",		8192	-    0,		256,	512		},
	{ "ATmega8515",		8192	- 2048,		512,	512		},
	{ "ATmega8515L",	8192	- 2048,		512,	512		},
	{ "ATmega8535",		8192	- 2048,		512,	512		},
	{ "ATmega8535L",	8192	- 2048,		512,	512		},
	{ "ATmega8",		8192	- 2048,		1024,	512		},
	{ "ATmega8L",		8192	- 2048,		1024,	512		},
	{ "ATmega16",		16384	- 2048,		1024,	512		},
	{ "ATmega16L",		16384	- 2048,		1024,	512		},
	{ "ATmega161",		16384	- 1024,		1024,	512		},
	{ "ATmega161L",		16384	- 1024,		1024,	512		},
	{ "ATmega162",		16384	- 2048,		1024,	512		},
	{ "ATmega162",		16384	- 2048,		1024,	512		},
	{ "ATmega162L",		16384	- 2048,		1024,	512		},
	{ "ATmega162V",		16384	- 2048,		1024,	512		},
	{ "ATmega163",		16384	- 2048,		1024,	512		},
	{ "ATmega163L",		16384	- 2048,		1024,	512		},
	{ "ATmega169",		16384	- 2048,		1024,	512		},
	{ "ATmega169L",		16384	- 2048,		1024,	512		},
	{ "ATmega169V",		16384	- 2048,		1024,	512		},
	{ "ATmega32",		32768	- 4096,		2048,	1024	},
	{ "ATmega32L",		32768	- 4096,		2048,	1024	},
	{ "ATmega323",		32768	- 4096,		2048,	1024	},
	{ "ATmega323L",		32768	- 4096,		2048,	1024	},
	{ "ATmega64",		65536	- 8192,		4096,	2048	},
	{ "ATmega64L",		65536	- 8192,		4096,	2048	},
	{ "ATmega103",		131072	-    0,		4096,	4096	},
	{ "ATmega103L",		131072	-    0,		4096,	4096	},
	{ "ATmega128",		131072	- 8192,		4096,	4096	},
	{ "ATmega128L",		131072	- 8192,		4096,	4096	},
	{ 0,				0		- 	 0,		0,		0 		}
};


// ------------------------------------------------------------------
// Global variables
// ------------------------------------------------------------------

void					*file_buffer;			// The entire contents of the ELF file
unsigned int			file_size;				// Size of the ELF file

Elf32_Header 			*elf_header;			// Pointer to file header
Elf32_Section_Header	*elf_section_headers;	// Pointer to beginning of section headers



// ------------------------------------------------------------------
//
// load_file()
//
// Loads a user-specified ELF file in memory. Verifies the file is
// an ELF file using the ELF magic numbers. Calculate pointers to
// data structures in the file to be used later in the application.
//
// Modifies globals:
//		file_buffer, file_size,
//		elf_header, elf_section_headers, elf_string_section
//
// ------------------------------------------------------------------
void load_file( char filename[] )
{
	FILE 		*file_stream;
	struct stat	file_status;

	// Open the file
	file_stream = fopen( filename, "rb" );
	if ( !file_stream )
	{
		printf( "avr-sizex: Unable to open file '%s'.\n", filename );
		exit( 1 );
	}

	// Determine the size of the file
	// Load the entire file into memory because AVR binaries are so small
	// Ensure the file is large enough to at least contain the header
    fstat( fileno(file_stream), &file_status );
	file_size = file_status.st_size;
	if ( file_size <= sizeof(Elf32_Header) )
	{
		printf( "avr-sizex: '%s' is not large enough to be an ELF file.\n", filename );
		exit( 1 );
	}

	// Allocate a buffer to hold the file
	// Read the entire file into file_buffer
	file_buffer = realloc( file_buffer, file_size );
    memset( file_buffer, 0x00, file_size );
    fread( file_buffer, file_size, 1, file_stream );
    fclose( file_stream );

	// The file should begin with the ELF header
	elf_header = file_buffer;

	// Ensure the file begins with the ELF magic number
	if ( memcmp(elf_header->e_ident, ELF_IDENT, sizeof(ELF_IDENT)) != 0 )
	{
		printf("avr-sizex: '%s' is not an ELF file (incorrect ELF ident).\n", filename );
		exit( 2 );
	}

	// Setup pointer to section headers
	elf_section_headers =  file_buffer + elf_header->e_shoff;

}

// ------------------------------------------------------------------
//
// get_elf_string()
//
// Retrieves a string located in the specified section.
//
// ------------------------------------------------------------------
char *get_elf_string(int string_offset, int section_number)
{
	Elf32_Section_Header	*header;
	char 					*elf_string_section;

	// Setup pointer to strings section
	// The strings are stored in their own section, there may be several string sections
	// Each string section has its own header listed along with all the other headers
	// Consider the section headers to be an array
	// The section header for "section_number" is section_headers[section_number]
	// Now, the strings are stored in the data of the section
	// The file offset for the section's data is at section_header.sh_offset
	// With this section data we can get section header names by adding string_offset
	header = elf_section_headers;
	header += section_number;						// jump ahead n section headers
	elf_string_section = file_buffer + header->sh_offset;
	return( elf_string_section + string_offset );
}

// ------------------------------------------------------------------
//
// find_section()
//
// Locates a section header structure with the specified text name
// and returns the section header structure to the caller. Returns
// NULL if section cannot be found. This function is
// case-insensitive.
//
// Uses globals:
//		elf_section_headers, elf_string_section
//
// ------------------------------------------------------------------
Elf32_Section_Header *find_section( char target_name[] )
{
	Elf32_Section_Header 	*header;
	char					*header_name;
	int i;

	// The first section header
	header = elf_section_headers;

	// Loop through the sections looking for target section
	for ( i=0; i < elf_header->e_shnum; i++ )
	{

		// Retrieve the name of this section
		header_name = get_elf_string(header->sh_name, elf_header->e_shstrndx);

		// Case insensitive compare
		if ( stricmp(header_name, target_name)==0 )
			return( header );

		// Jump to the next header
		header++;
	}

	return( NULL );
}


// ------------------------------------------------------------------
//
// get_section_size()
//
// Retrieves the size in bytes of a section's data. The size of a
// section's data is stored in the section header structure. This
// function is case-insensitive.
//
// ------------------------------------------------------------------
unsigned int get_section_size( char target_name[] )
{
	Elf32_Section_Header *header;

	// Find the section header by name
	header = find_section(target_name);
	if ( header==NULL )
	{
		printf("avr-sizex: Unable to find the %s section.\n", target_name);
		exit( 3 );
	}

	// Return the section's size in bytes
	return( header->sh_size );
}

// ------------------------------------------------------------------
//
// find_device()
//
// Locates a Atmel AVR device from the internal (hard-coded) list of
// microcontrollers in the devices[] array. Returns the index into
// the array of the device requested or -1 of the device cannot be
// found. This function is case-insensitive.
//
// Uses globals
//		devices[]
// ------------------------------------------------------------------
int find_device( char target_name[] )
{
	int c = 0;

	while ( devices[c].name != NULL )
	{
		if ( stricmp(devices[c].name, target_name)==0 )
			return( c );

		c++;
	}

	return( -1 );
}


// ------------------------------------------------------------------
//
// print_usage()
//
// Print to console instructions on using this utility.
//
// ------------------------------------------------------------------
void print_usage()
{
	printf("avr-sizex v%s \nDisplay the percentage of space used on an AVR device.\n", VERSION);
	printf("\n");
	printf("Usage: avr-sizex [options] file.elf\n");
	printf("Options:\n");
	printf("  -m --mcu [avr-name]     Specify the AVR microcontroller name\n");
	printf("\n");
	printf("  -f --flash [size]       Specify the size of flash memory in bytes\n");
	printf("  -s --sram [size]        Specify the size of sram memory in bytes\n");
	printf("  -e --eeprom [size]      Specify the size of eepprom memory in bytes\n");
	printf("\n");
	printf("  -l --long               Long listing; display addition information\n");
	printf("  -t --top [n]            List the N largest objects by size\n");
	printf("     --nowarn             Suppress warning messages when device is nearly full\n");
	printf("     --noerror            Suppress error messages when device is over capacity\n");
	printf("  -v --version            Display program version number\n");
	printf("  -h --help               Display this information\n");
	printf("  -d --devices            Display a list of known devices\n");
	printf("\n");
	printf("Example:\n");
	printf("  avr-sizex --mcu atmega8 myprogram.elf\n");
	printf("  avr-sizex --flash 2048 --sram 512 --eeprom 128 myprogram.elf\n");
	printf("\n");
	printf("The --mcu option is required. If your avr device is not listed,\n");
	printf("  use the --flash and --sram options instead to specify the size\n");
	printf("  of the device.\n");
	printf("\n");
	printf("Bugs, comments, suggestions: brian.low@shaw.ca\n");

}

// ------------------------------------------------------------------
//
// print_version()
//
// Print to console the version number of this utility. See VERSION
// define at top of this file.
//
// ------------------------------------------------------------------
void print_version()
{
	printf("avr-sizex version %s\n", VERSION);
}

// ------------------------------------------------------------------
//
// print_devices()
//
// Print to console the list of devices this utility knows about.
// Includes device sizes.
//
// Uses globals:
//		devices[]
// ------------------------------------------------------------------
void print_devices()
{
	int c = 0;
	int col = 0;

	printf("%-15s %8s %8s %8s\n", "Name", "Flash", "SRAM", "EEPROM");
	printf("%-15s %8s %8s %8s\n", "----", "-----", "----", "------");

	while ( devices[c].name != NULL )
	{
		printf("%-15s %8i %8i %8i\n", devices[c].name, devices[c].flash, devices[c].sram, devices[c].eeprom);

		c++;
	}

	printf("\n");
	printf("* All sizes are in bytes.\n");
	printf("* The flash size has been adjusted down for all ATmega devices\n");
	printf("  with the self-programming feature. These devices have a reserved\n");
	printf("  boot area. The factory default size was used. If you have\n");
	printf("  changed the size of the boot area with the fuse bits, use\n");
	printf("  the --flash option to override the size of the flash memory.\n");
}


// ------------------------------------------------------------------
//
// print_topN()
//
// Print to console a list of the num_top largest functions
// and global/static variables
//
// Uses globals:
//		file_buffer
// ------------------------------------------------------------------

void print_topN(int num_top)
{
	Elf32_Section_Header 	*header;			// pointer to .symtab section header
	Elf32_Symbol 			*symbols;			// pointer to a symbol
	int						string_section;		// index to section with strings
	int 					num_symbols;		// number of symbols in elf file
	int 					i;					// temp index
	int						f;					// index into top_function
	int						v;					// index into top_variable
	int						x;					// temp index
	char					*s;					// temp string

	Elf32_Symbol			*top_functions[num_top];	// array of pointers to symbols
	Elf32_Symbol			*top_variables[num_top];	// array of pointers to symbols


	// Initialize the TopN list of functions
	for (f=0; f<num_top; f++)
		top_functions[f] = NULL;

	// Initialize the TopN list of variables
	for (v=0; v<num_top; v++)
		top_variables[v] = NULL;

	// Locate the symbol table
	header = find_section(".symtab");						// Find the .symtab section
	symbols = file_buffer + header->sh_offset;				// Find the .symtab section data
	string_section = header->sh_link;						// Find the # of the section holding the symbol strings
	num_symbols = header->sh_size / sizeof(Elf32_Symbol);	// Find the number of symbols stored

	// Loop through each symbol
	for(i=0; i<num_symbols; i++)
	{
		// Is this symbol a function?
		if (ELF32_ST_TYPE(symbols[i].st_info) == STT_FUNC)
		{
			// Loop through the existing topN list of functions
			// to see the current symbol should be inserted
			for (f=0; f<num_top; f++)
			{
				// Is the current topN spot empty,
				// or is the current symbol larger than the current topN spot
				if ( (top_functions[f]==NULL) || (symbols[i].st_size > top_functions[f]->st_size) )
				{
					// We are inserting the new symbol, this short loop
					// moves all subsequent topN spots down one to make room
					x = num_top-1;
					while (x>f)
					{
						top_functions[x] = top_functions[x-1];
						x--;
					}

					// Insert the new symbol
					top_functions[f] = &symbols[i];

					break;
				}
			}
		}

		// Is this symbol a variable (object)?
		if (ELF32_ST_TYPE(symbols[i].st_info) == STT_OBJECT)
		{
			// Loop through the existing topN list of variables
			// to see the current symbol should be inserted
			for (v=0; v<num_top; v++)
			{
				// Is the current topN spot empty,
				// or is the current symbol larger than the current topN spot
				if ( (top_variables[v]==NULL) || (symbols[i].st_size > top_variables[v]->st_size) )
				{
					// We are inserting the new symbol, this short loop
					// moves all subsequent topN spots down one to make room
					x = num_top-1;
					while (x>v)
					{
						top_variables[x] = top_variables[x-1];
						x--;
					}

					// Insert the new symbol
					top_variables[v] = &symbols[i];

					break;
				}
			}
		}

	}

	// Display results

	printf("Top Functions                        Top Variables (global/static) \n");
	printf("------------------------------       ------------------------------\n");

	for(i=0; i<num_top; i++)
	{
		if ( top_functions[i]!=NULL )
		{
			s = get_elf_string(top_functions[i]->st_name, string_section);
			printf("%-18.18s %5i bytes       ", s, top_functions[i]->st_size);
		}

		if ( top_variables[i]!=NULL )
		{
			s = get_elf_string(top_variables[i]->st_name, string_section);
			printf("%-18.18s %5i bytes", s, top_variables[i]->st_size);
		}

		if ( (top_functions[i]!=NULL) || (top_variables[i]!=NULL) )
			printf("\n");
	}

}


// ------------------------------------------------------------------
//
// main()
//
// ------------------------------------------------------------------
int main( int argc, char * argv[] )
{
	int	device_flash	= 0;		// number of bytes of available flash memory
	int	device_sram		= 0;		// number of bytes of available sram memory
	int	device_eeprom   = 0;		// number of bytes of available eeprom memory

	char no_warn = 0;				// suppress warning message when near memory limit
	char no_error = 0;				// suppress error messages when over memory limit
	char long_listing = 0;			// long listing; display addition info

	int num_top = -1;				// number of functions/variables to display on te topN lists

	// -----------------  Process options  -----------------

	int i;							// temp
	char ch;						// selected option
	char *success;					// temp
	while ( (ch = getopt_long_only(argc, argv, "m:f:s:e:t:vthl?", longopts, NULL)) != -1 )
	{
		switch (ch)
		{
			case 'm':
				i = find_device( optarg );
				if ( i >= 0 )
				{
					device_flash = devices[i].flash;
					device_sram = devices[i].sram;
					device_eeprom = devices[i].eeprom;
				}
				else
				{
					printf("avr-sizex: The device '%s' is not known.\n", optarg);
					printf("  Use --flash, --sram and --eeprom instead or,\n");
					printf("  use --devices to see a list of known devices.\n");
					exit( 1 );
				}
				break;

			case 'f':
				device_flash = strtol( optarg, &success, 0 );
				if ( *success != '\0' )
				{
					printf("avr-sizex: Please specify a valid number after the --flash option\n");
					exit(5);
				}
				break;

			case 's':
				device_sram = strtol( optarg, &success, 0 );
				if ( *success != '\0' )
				{
					printf("avr-sizex: Please specify a valid number after the --sram option\n");
					exit(5);
				}
				break;

			case 'e':
				device_eeprom = strtol( optarg, &success, 0 );
				if ( *success != '\0' )
				{
					printf("avr-sizex: Please specify a valid number after the --eeprom option\n");
					exit(5);
				}
				break;

			case 'w':
				no_warn = 1;
				break;

			case 'o':
				no_error = 1;
				break;

			case 'l':
				long_listing = 1;
				break;

			case 't':
				num_top = 5;
				if ( optarg != NULL )
				{
					num_top = strtol( optarg, &success, 0 );
					if ( *success != '\0' )
					{
						printf("avr-sizex: Please specify a valid number after the --top option\n");
						exit( 5 );
					}
				}
				break;

			case '?':
			case 'h':
				print_usage();
				exit( 0 );
				break;

			case 'v':
				print_version();
				exit( 0 );
				break;

			case 'd':
				print_devices();
				exit( 0 );
				break;

			default:
				print_usage();
				exit( 10 );
				break;

		}

	}

	// If no arguments specified, display the usage
	if ( argc <= 1 )
	{
		print_usage();
		exit( 1 );
	}

	// Ensure that device flash size has been set
	if ( (device_flash <= 0) )
	{
		printf("avr-sizex: Please specify the size of the the AVR device with either the \n");
		printf("  --mcu option or a combination of the --flash, --sram and --eeprom options.\n");
		printf("  See --help for more information.\n");
		exit( 1 );
	}

	// Ensure a file has been specified
	if ( optind > argc-1 )
	{
		printf("avr-sizex: No input file. See --help for more information.\n");
		exit( 1 );
	}



	// -----------------  Load file and calculate sizes  -----------------

	// Load the file into memory
	// Setup global pointers to important data structures
	load_file(argv[optind]);

	// Get the sizes of the sections to calculate code size
	int text_size	= get_section_size(".text");
	int data_size	= get_section_size(".data");
	int bss_size 	= get_section_size(".bss");
	int eeprom_size	= get_section_size(".eeprom");

	// Calculate the space used on the device
	int flash_used = text_size + data_size;
	int sram_used = data_size + bss_size;
	int eeprom_used = eeprom_size;

	// Calculate percentage
	double flash_percent = 	(device_flash==0 ?	0.0 : 100 * flash_used/device_flash);
	double sram_percent = 		(device_sram==0 ?	0.0 : 100 * sram_used/device_sram);
	double eeprom_percent = 	(device_eeprom==0 ?	0.0 : 100 * eeprom_used/device_eeprom);




	// -----------------  Display results  -----------------

	// Display results
	printf("\n");
	printf("    Flash     SRAM     EEPROM \n");
	printf("    -----     ----     ------ \n");
	printf("  %6.0f%%  %6.0f%%    %6.0f%%\n", flash_percent, sram_percent, eeprom_percent);

	if ( long_listing )
	{
		printf("   %6i   %6i     %6i    bytes used\n", flash_used, sram_used, eeprom_used);
		printf("   %6i   %6i     %6i    bytes free\n", device_flash-flash_used, device_sram-sram_used, device_eeprom-eeprom_used);
		printf("   %6i   %6i     %6i    bytes capacity\n", device_flash, device_sram, device_eeprom);
	}




	// -----------------  Display size errors and warnings  -----------------

	if ( flash_percent > 100 ) {
		if ( !no_error ) {
			printf("\nERROR: You have exceeded the flash memory available on this device.\n       Reduce code size.\n       Reduce pre-initialized global and static variables.\n       Reduce the size of the boot area with the fuse bits (ATmega).\n");
		}
	}
	else if ( flash_percent > 90 ) {
		if ( !no_warn ) {
			printf("\nWarning: You have nearly exceeded the flash memory available on this device.\n         Reduce code size.\n         Reduce pre-initialized global and static variables.\n         Reduce the size of the boot area with the fuse bits (ATmega).\n");
		}
	}

	if ( sram_percent > 100 ) {
		if ( !no_error ) {
			printf("\nERROR: You have exceeded the SRAM available on this device.\n       Free SRAM is required for the stack!\n       Reduce depth of nested function calls.\n       Reduce global, static and local variables.\n");
		}
	}
	else if (sram_percent > 90) {
		if ( !no_warn ) {
			printf("\nWarning: You have nearly exceeded the SRAM available on this device.\n         Free SRAM is required for the stack!\n         Reduce depth of nested function calls.\n         Reduce global, static and local variables.\n");
		}
	}

	if ( eeprom_percent > 100 ) {
		if ( !no_error ) {
			printf("\nERROR: You have exceeded the EEPROM memory available on this device.\n");
		}
	}
	else if (eeprom_percent > 90) {
		if ( !no_warn ) {
			printf("\nWarning: You have nearly exceeded the EEPROM memory available on this device.\n");
		}
	}

	printf("\n");


	// -----------------  Display TopN Lists  -----------------

	if (num_top > 0)
		print_topN(num_top);



	// -----------------  Cleanup  -----------------

	free( file_buffer );
	exit( 0 );
}

